"""Script to load reach CrossSection Information.
# The program can be executed like so: 'python -m entfrachten.analysis.cross_section_info_loader some_folder/some_model.sqlite -a "\\.\pipe\some_pipe"'
Where some_model is the mike model you would like to read from.
# """

def start_listener():
    """Start the listener. Note that this also executes imports which may prevent you from using mikeio1d."""
    # imports are inside the function because they have side effects
    import argparse
    from multiprocessing.connection import Listener
    from warnings import warn

    import numpy as np
    import scipy.integrate

    # mike api loading, its a c# api, clr.AddReference is used for loading
    # first import clr
    import clr

    # use AddReference to add a reference to the dll (in c# this would be a using directive).
    # DHI.Mike.Install needs some additional parameters, I have copied those from the python examples from DHI.
    # see https://github.com/DHI/Mike1D-SDK-examples/blob/master/scripts/ResultDataExtract.py
    clr.AddReference(
        "DHI.Mike.Install, Version=1.0.0.0, Culture=neutral, PublicKeyToken=c513450b5d0bf0bf"
    )
    # after using AddReference, we can import the c# libraries like python modules.
    from DHI.Mike.Install import MikeImport, MikeProducts

    MikeImport.SetupLatest(MikeProducts.MikePlus)
    clr.AddReference("DHI.Mike1D.Generic")
    from DHI.Mike1D.Generic import (
        Connection,
        Diagnostics,
        Location,
        LocationSpan,
        XYZLocationSpan,
        ZLocation,
    )

    clr.AddReference("DHI.Mike1D.Mike1DDataAccess")
    from DHI.Mike1D.Mike1DDataAccess import Mike1DBridge

    clr.AddReference("DHI.Mike1D.NetworkDataAccess")
    from DHI.Mike1D.NetworkDataAccess import IReach

    clr.AddReference("DHI.Mike1D.CrossSectionModule")
    from DHI.Mike1D.CrossSectionModule import ICrossSection, CrossSectionData, ICrossSection

    # ArgumentParser helps to add command line arguments.
    # Just use .add_argument and it takes care of parsing the command line input.
    parser = argparse.ArgumentParser("cross_section_info_loader")
    parser.add_argument("model_path", help="path of mike model", type=str)
    parser.add_argument(
        "-a",
        "--address",
        default=R"\\.\pipe\xs_pipe",
        help="address of listener, client needs to use the same",
        type=str,
    )
    # this is where the parser does the actual parsing, raises an exception and prints help info if it fails.
    args = parser.parse_args()

    # we do not have topos, but have to pass an argument, which needs to be an empty string.
    NULL_TOPO = ""

    # Load the model data.
    print("loading file ", args.model_path)
    # Many methods from the mike1d api need this diagnostics object as an input. Probably used for error messages.
    diagnostics = Diagnostics("diagnostics")
    # the connection tells the api where to read the model data from
    connection = Connection.Create(args.model_path)
    # May need a simulation id if there are multiple scenarios:
    # connection.Options.Add("simulationId", "397")
    # The bridge helps to read the model data from any supported format
    mike1DBridge = Mike1DBridge()
    mike1DData = mike1DBridge.Open(connection, diagnostics)
    # reaches are contained as list, put into dict with name as key
    reach_dict: dict[str, IReach] = {
        r.LocationSpan.ID: r for r in mike1DData.Network.Reaches
    }
    # this is the part of the data we need
    csData = mike1DData.CrossSections

    # use the Listener to communicate with another process.
    address = args.address
    if address.startswith("(") and address.endswith(")"):
        print("trying to parse port from {address}")
        port = int(address.split(",")[-1].removesuffix(")"))
        print(f"set port to {port}")
        address = ("localhost", port)
        print(f"set address to {address}")
    print(f"starting server at {address} with authkey b'reach_reader'")
    listener = Listener(address, authkey=b"reach_reader")
    print(f"started server at {address} with authkey b'reach_reader'")
    conn = listener.accept()
    print("connection accepted from", listener.last_accepted)
    should_run = True
    while should_run:
        try:
            msg = conn.recv()
            # otherwise assume its a reach id
            if isinstance(msg, tuple):
                fkt = msg[0]
                if fkt == "full_xs_area":
                    if not len(msg) == 2:
                        raise ValueError(
                            "Could not parse. To get full cross section area for a reach send ('full_xs_area',reach)."
                        )
                    reach = msg[1]
                    # first argument is the reach name. Second is topo_id, which is not used in our case
                    cross_section = csData.FindGlobalCrossSection(reach, NULL_TOPO)
                    # need to call initialize before the cross section can give us values.
                    cross_section.Initialize()
                    # send data back to client
                    conn.send(
                        cross_section.GetArea(
                            cross_section.BottomLevel + cross_section.Height
                        )
                    )
                elif fkt == "xs_areas_at":
                    if not len(msg) == 4:
                        raise ValueError(
                            "Could not parse. To get cross section area for a reach at specific waterlevel and chainages send ('xs_areas_at', reach, water_level, chainage_list)."
                        )
                    reach = msg[1]
                    reach_span: XYZLocationSpan = reach_dict[reach].LocationSpan
                    water_level = float(msg[2])
                    chainages = np.array(msg[3]).clip(0, reach_span.EndChainage)
                    cross_areas = np.full_like(chainages, np.nan)
                    for i, c in enumerate(chainages):
                        # GetCrossSection does interpolation, this should take care of changing reach heights or cross_section shapes.
                        # This requires a location with Z-Value, which we can get from the LocationSpan of the IReach object.
                        loc = reach_span.CreateLocation(c)
                        cross_section = csData.GetCrossSection(loc, NULL_TOPO)
                        if not cross_section is None:
                            # need to call initialize before the cross section can give us values.
                            cross_section.Initialize()
                            # might raise an exception if the water level is too high, clip.
                            effective_water_level = np.clip(
                                water_level,
                                cross_section.BottomLevel,
                                cross_section.BottomLevel + cross_section.Height,
                            )
                            cross_areas[i] = cross_section.GetArea(effective_water_level)
                    conn.send((chainages, cross_areas))
                elif fkt == "volume_from_waterlevels":
                    if not len(msg) == 4:
                        raise ValueError(
                            "Could not parse. To get volume in a reach from water levels send ('volume_from_waterlevels', reach, chainages ,water_levels)."
                        )
                    reach = msg[1]
                    chainages = msg[2]
                    water_levels = msg[3]
                    if not len(chainages) == len(water_levels):
                        raise ValueError(
                            "chainages and water levels need to have same length"
                        )
                    reach_span: XYZLocationSpan = reach_dict[reach].LocationSpan

                    cross_areas = np.full_like(chainages, np.nan)
                    for i, c in enumerate(chainages):
                        # GetCrossSection does interpolation, this should take care of changing reach heights or cross_section shapes.
                        # This requires a location with Z-Value, which we can get from the LocationSpan of the IReach object.
                        loc = reach_span.CreateLocation(c)
                        cross_section = csData.GetCrossSection(loc, NULL_TOPO)
                        if not cross_section is None:
                            # need to call initialize before the cross section can give us values.
                            cross_section.Initialize()
                            # might raise an exception if the water level is too high, clip.
                            effective_water_level = np.clip(
                                water_levels[i],
                                cross_section.BottomLevel,
                                cross_section.BottomLevel + cross_section.Height,
                            )
                            cross_areas[i] = cross_section.GetArea(effective_water_level)
                    volume = scipy.integrate.trapezoid(y=cross_areas, x=chainages)
                    conn.send(volume)
            # close if msg is "close"
            if msg == "close":
                print("closing now")
                should_run = False
        except Exception as e:
            print("Exception occurred.")
            warn(str(e))
            # send a new Exception made with just the string, this avoids errors where the original exception cannot be pickled
            conn.send(Exception(str(e)))
    conn.close()
    listener.close()

if __name__ == "__main__":
    start_listener()
