"""Module for graph problems on the sewer network. Names are the same as in the mike model:
Nodes may be manholes or basins or pipe junctions.
Any connection between the nodes are called reaches, reaches are usually pipes but may also be weirs or pumps.
So reaches are what would be called edges in graph theory."""

import collections
import typing
import pandas as pd

class PathNotFoundException(Exception):
    pass

class DrainageSystemGraph:
    def __init__(self,reach_df:pd.DataFrame) -> None:
        """reach_df: index is the name of the reach, needs columns 'start' and 'end' which define which nodes the reach connects."""
        self.reach_df = reach_df

    def find_upstream_nodes_reaches(self,start:str, stop_list:typing.Iterable[str]) -> typing.Tuple[set[str],set[str]]:
        """finds all nodes and reaches upstream of start, where the connection does not go through a node from stop_list.
        Return a set of nodes and a set of reaches."""
        # set is better at membership tests
        stop_set = set(stop_list)
        q = collections.deque([start])
        found_nodes = {start}
        found_reaches = set()
        while len(q) > 0:
            current_node = q.popleft()
            # query for reaches that end at current_node.
            # checks for q or stop_list has to be done on the node,
            # because the reaches have to be added, even if the node should not be added to the list,
            reaches = self.reach_df.query(
                "end == @current_node "
            )
            for reach_name in reaches.index:
                found_reaches.add(reach_name)
            for next_node in reaches["start"]:
                if not next_node in stop_set or next_node in found_nodes:
                    found_nodes.add(next_node)
                    q.append(next_node)
        return found_nodes, found_reaches



    def find_shortest_path_reach(self,start_reach:str, end_reach:str):
        """Find the shortest path (in number of links!) from start to end. 
        Return the path as a list of reaches"""
        # Standard breadth first search.
        # Since we only care about reaches, those are treated as the nodes of the graph,
        # the edges are reach-pairs where start and end match.
        q = collections.deque([start_reach])
        explored = {start_reach}
        parents = {}
        while len(q) > 0:
            current_reach = q.popleft()
            if current_reach == end_reach:
                break
            # Note that this variable is used in query()
            current_end_node = self.reach_df.loc[current_reach, "end"]
            for next_reach in self.reach_df.query("start==@current_end_node").index:
                if not next_reach in explored:
                    explored.add(next_reach)
                    parents[next_reach] = current_reach
                    q.append(next_reach)
        if current_reach != end_reach:
            raise PathNotFoundException(f"No path found from {start} to {end}")
        path = [current_reach]
        while path[-1] != start_reach:
            path.append(parents[path[-1]])
        return list(reversed(path))
    

    def find_shortest_path(self, start:str, end:str):
        """Find the shortest path (in number of links!) from start to end. 
        Return the path as a list of reaches"""
        # Standard breadth first search.
        q = collections.deque([start])
        # set of explored nodes, those are not visited again
        explored = {start}
        # for each visited node, save the reach from which the node was visited.
        # this is later used to follow the path back
        to_parent = {}
        while len(q) > 0:
            current_node = q.popleft()
            if current_node == end:
                break
            for reach, next_node in self.reach_df.query("start==@current_node")["end"].items():
                if not next_node in explored:
                    explored.add(next_node)
                    to_parent[next_node] = reach
                    q.append(next_node)
        if current_node != end:
            raise PathNotFoundException(f"No path found from {start} to {end}")
        reach = to_parent[current_node]
        reach_start = self.reach_df.loc[reach,"start"]
        path = [reach]
        while  reach_start != start:
            reach = to_parent[reach_start]
            reach_start = self.reach_df.loc[reach,"start"]
            path.append(reach)
        return list(reversed(path))

    def find_next_upstream_nodes(self,subset:typing.Iterable[str]) -> typing.Dict[str,set[str]]:
        "For each node in subset find the other nodes from subset that can be reached upstream, without going over another node from subset."
        # set is better for member checks
        subset = set(subset)
        next_subset_nodes = {}
        for s in subset:
            next_subset_nodes[s] = set()
            q = collections.deque([s])
            explored = {s}
            while len(q) > 0:
                # Note that this variable is used in query()
                current_node = q.popleft()
                for next_node in self.reach_df.query("end==@current_node and not start in @explored")["start"]:
                    explored.add(next_node)
                    if next_node in subset:
                        # add to next_structures. 
                        # do not add to q, so that exploration stops here.
                        next_subset_nodes[s].add(next_node)
                    else:
                        q.append(next_node)
        return next_subset_nodes