"""This module defines multiple loss functions which may be used to compare the simulation results to the reference.
Note that nevergrad performs minimization, therefore smaller values need to indicate a better result."""
from warnings import warn

import pandas as pd
import numpy as np
import hydroeval as hde

def rmse(ref: pd.Series, res: pd.Series) -> float:
    """Root Mean Square Error.
    No weighting or averaging over labels is applied."""
    if not ref.index.equals(res.index):
        raise ValueError("Indices are not equal, cannot compare")
    return ((ref - res) ** 2).mean(axis=None) ** 0.5


def _vfp(ref: pd.Series, res: pd.Series, divisor_on_zero) -> float:
    """Difference of total volumes as a percentage."""
    if not ref.index.equals(res.index):
        raise ValueError("Indices are not equal, cannot compare")
    # Total reference Volume
    ref_volume = ref.sum()
    # Total result Volume
    res_volume = res.sum()
    dif = abs((res_volume - ref_volume))
    # if reference volume is 0 use the replacement value this is more useful for optimization than nan or zero.
    if ref_volume == 0:
        ref_volume = divisor_on_zero
    # Convert to percent
    return 100 * dif / ref_volume

def weighted_vfp_factory(result_spec,reference):
    """From result spec and reference create a function that applies vfp to each label
    and then takes the weighted mean.
    """
    # for vfp: replacement value for the divisor when ref_volume is zero.
    # it may be better manually define one for each label.
    divisor_on_zero = reference.groupby(level="Label").sum().mean()
    def weighted_vfp(ref: pd.Series, res: pd.Series) -> float:
        """Applies vfp to each label and then takes the weighted mean."""
        labels = result_spec.index.get_level_values("Label")
        # also put timestep into weights, since vfp does not account for time passed
        weights = (
            result_spec["EventWeight"].to_numpy()
            * pd.to_timedelta(result_spec["Timestep"]).dt.total_seconds().to_numpy()
        )
        per_label = np.array(
            [_vfp(ref.xs(l, level="Label"), res.xs(l, level="Label"),divisor_on_zero) for l in labels]
        )
        nan_parts = np.isnan(weights) | np.isnan(per_label)
        if np.any(nan_parts):
            warn("weighted_vfp: some parts are nan, skipping those")
        return np.average(per_label[~nan_parts], weights=weights[~nan_parts])
    return weighted_vfp


# note: if your objective function should be maximized, use the negative objective function as loss function
def mean_negative_bounded_nse(ref: pd.Series, res: pd.Series) -> float:
    """
    Mean over labels of negative nse_c2m. range [-1,1].
    nse_c2m is very similar to NSE, but bounded to range[-1,1], which is better for optimization.
    Negation is applied so that better values are smaller.
    Args:
        actual (pandas.DataFrame): DataFrame containing actual values.
        forecast (pandas.DataFrame): DataFrame containing forecasted values.
    Returns:
        float: Cost of the solution.
    """
    if not ref.index.equals(res.index):
        raise ValueError("Indices are not equal, cannot compare")
    loss_parts = []
    for label in ref.index.unique("Label"):
        ref_part = ref.xs(label, level="Label")
        res_part = res.xs(label, level="Label")
        if ref_part.isna().any() or res_part.isna().any():
            warn(f"Label {label} ref or res has nan values")
            ref_part = ref_part.interpolate().ffill().bfill().fillna(0)
            res_part = res_part.interpolate().ffill().bfill().fillna(0)
        loss_parts.append(-hde.nse_c2m(res_part.to_numpy(), ref_part.to_numpy()))
    return np.mean(loss_parts)