
from pathlib import Path
from warnings import warn
import dill


class OptimDumper:
    """Class for dumping and loading the optimization state of a nevergrad optimizer.
    New files are created for each step, the old files are kept as a back up.
    For a large optimization budget dump should not be called every step to reduce storage use.
    """

    _fin_format = "{0}.fin"

    def __init__(self, dump_folder):
        """folder_name: folder (relative to temp_folder) where dumps are stored."""
        self.dump_folder = Path(dump_folder)
        self.dump_folder.mkdir(exist_ok=True)

    # the following methods define the path where the file for a variable and step should be stored.

    def _get_path_opt(self, step):
        return self.dump_folder / f"{step}.opt.dill"

    def _get_path_params(self, step):
        return self.dump_folder / f"{step}.par.dill"

    def _get_path_losses(self, step):
        return self.dump_folder / f"{step}.los.dill"

    def _get_path_fin(self, step):
        return self.dump_folder / self._fin_format.format(step)

    def get_last_step(self) -> int:
        """Returns the last step for which a dump was finished."""
        # {n}.fin file is used to mark successful dump, where n is the optimization step.
        # The .fin file ensures that we can easily find tha last valid dump, even if the next dump was interrupted.
        fin_steps = [
            int(f.stem) for f in self.dump_folder.glob(self._fin_format.format("*"))
        ]
        if len(fin_steps) == 0:
            return -1
        return max(fin_steps)

    def load(self):
        """Load the last optimizer state.
        Returns: optimizer, list of asked parameters and list of losses."""
        last_step = self.get_last_step()
        if last_step < 0:
            raise DumpNotFoundException(f"No dump was finished for {self.dump_folder}")
        with open(self._get_path_opt(last_step), "rb") as fo:
            optimizer = dill.load(fo)
        if optimizer.num_tell != (last_step + 1):
            warn("optimizer.num_tell does not match detected last step")
        print(
            f"loaded optimizer from {self._get_path_opt(last_step)}. executed iterations: {optimizer.num_tell}"
        )
        with open(self._get_path_params(last_step), "rb") as fp:
            params = dill.load(fp)
        if optimizer.num_tell != len(params):
            warn("length of params does not match")
        with open(self._get_path_losses(last_step), "rb") as fl:
            losses = dill.load(fl)
        if optimizer.num_tell != len(losses):
            warn("length of losses does not match")
        return optimizer, params, losses

    def dump(self, optimizer, params, losses):
        """Save optimization variables, which cannot be easily recreated, to disk."""
        step = optimizer.num_tell - 1
        with open(self._get_path_opt(step), "wb") as f:
            dill.dump(optimizer, f)
        with open(self._get_path_params(step), "wb") as f:
            dill.dump(params, f)
        with open(self._get_path_losses(step), "wb") as f:
            dill.dump(losses, f)
        # create file to mark end of step. only the name matters, thus do not write anything into the file.
        open(self._get_path_fin(step), "w").close()


class DumpNotFoundException(Exception):
    pass
