"""Provides a class for setting parameters and running a simulation with mikepluspy.
"""

import os
import typing
from contextlib import redirect_stdout
from warnings import warn
from pathlib import Path
import threading
from multiprocessing.pool import ThreadPool

import pandas as pd
from mikeplus import DataTableAccess
from mikeplus.engines import Engine1D
from mikeio1d.res1d import Res1D

# delete  result files(as specified by event config) before a sim, assures that no old file is reread in the case that a simulation does not run
DELETE_RESFILE_BEFORE_SIM = True
# For dates the dta returns a .NET Datetime, which pandas cannot convert. Use ToString first.
DOTNET_DT_ISO_FORMAT = "yyyy-MM-ddTHH:mm:ss"


class SimulationRunner:
    def __init__(
        self,
        db_path: typing.Union[str, Path],
        result_folder: typing.Union[str, Path],
        param_spec: pd.DataFrame,
        result_spec: pd.DataFrame,
        parallel_limit: int,
        adjust_end_time: bool = True,
    ) -> None:
        self.result_folder = Path(result_folder)
        self.param_spec = param_spec
        if not {
            "ResultFile",
            "ResultReach",
            "ResultNode",
            "ResultColumn",
            "Timestep",
            "Start",
            "End",
        }.issubset(result_spec.columns):
            raise ValueError(
                "result_spec needs columns: ResultFile,ResultReach,ResultNode,ResultColumn,Timestep,Start,End"
            )
        if not (
            result_spec["Start"].dt.tz is None and result_spec["End"].dt.tz is None
        ):
            raise ValueError(
                "Timezone for Start and End has to be None. Use dt.tz_localize(None) to drop timezone."
            )

        # Preparations for running the simulation
        # open the MIKE+ database using MIKE+Py.
        self.result_spec = result_spec
        self.dta = DataTableAccess(db_path)
        self.dta.open_database()
        assert self.dta.is_database_open()
        # Dict Engine1D objects that will be used to run MIKE 1D simulations, one per thread
        self.engines = {}

        def init_engine():
            self.engines[threading.current_thread()] = Engine1D(self.dta.datatables)
            print(f"initialized engine for {threading.current_thread()}")

        # sim_executor is used to run multiple simulations in parallel.
        # Since the actual simulation runs in its own process anyway, it makes no sense to use multiple python processes.
        # Threadpool from multiprocessing is used, because it has a rather simple interface.
        self.sim_executor = ThreadPool(
            processes=parallel_limit, initializer=init_engine
        )
        if adjust_end_time:
            self._adjust_computation_end()
        self.sim_longest_first = self._get_sim_ordered_by_duration()

    def _adjust_computation_end(self):
        """Adjust computation times so that nothing unnecessary is computed"""
        sim_group = self.result_spec.reset_index().groupby("SimulationName")
        # these define the time span we need to read from
        read_end = sim_group["End"].max()
        sim_pad = pd.to_timedelta("5min")
        for sim in read_end.index:

            # Adjust end. Use string to set value, since a Pandas datetime cannot be converted.
            end_string = (read_end[sim] + sim_pad).isoformat()
            self.dta.set_value("msm_Project", sim, "ComputationEnd", end_string)
            # make sure the result matches
            assert (
                self.dta.get_field_values("msm_Project", sim, ["ComputationEnd"])[
                    0
                ].ToString(DOTNET_DT_ISO_FORMAT)
                == end_string
            )

    def _get_sim_ordered_by_duration(self):
        """Make a list of simulations where the longest comes first.
        Assuming that longer simulations use more computation time,
        this prevents a case where a long simulation runs at the end and takes up only on thread,
        while other threads are idle."""
        sim_group = self.result_spec.reset_index().groupby("SimulationName")
        # start of reading time, checked later
        read_start = sim_group["Start"].min()
        duration = pd.Series(data=pd.to_timedelta("NaT"), index=read_start.index)
        for sim in duration.index:
            # read start
            computation_start = pd.to_datetime(
                self.dta.get_field_values("msm_Project", sim, ["ComputationBegin"])[
                    0
                ].ToString(DOTNET_DT_ISO_FORMAT)
            )
            # check start time
            if computation_start > read_start[sim]:
                warn(
                    f"Simulation {sim} start too late. Got {computation_start} need {read_start[sim]}"
                )
            # read end
            computation_end = pd.to_datetime(
                self.dta.get_field_values("msm_Project", sim, ["ComputationEnd"])[
                    0
                ].ToString(DOTNET_DT_ISO_FORMAT)
            )
            duration[sim] = computation_end - computation_start
        return duration.sort_values(ascending=False).index.to_list()

    def close(self):
        self.dta.close_database()
        self.sim_executor.close()

    def _single_sim_run(self, sim_name: str):
        """Runs a single simulation for event ev. Returns a string message on successful finish.
        Intended for usage with ThreadPool.map(). Use map to complete all simulations and read results after that
        """
        # also measure the time the simulation takes
        start_time = pd.Timestamp.now()
        print(f"Running simulation {sim_name} in {threading.current_thread()}.")
        self.engines[threading.current_thread()].run(sim_name)
        end_time = pd.Timestamp.now()
        return f"Completed sim for {sim_name},took {end_time-start_time},in {threading.current_thread()}"

    # the ** means that simulation run may be called like so:
    # result = simulation_run(key1=p1,key2=p2, ...)
    # but kwparams will contain a dict of form {"key1":p1,"key2":p2,...}
    # if you called the simulation like simulation_run(key1=p1,key2=p2, ...),
    # key1, key2, ... would need to be valid python names
    # but with simulation_run(**{"key1":p1,"key2":p2,...}) key1, key2 could be any string.
    # Because the optimizer uses the second form, we can use the index from param_specs as a key,
    # even though the index may not be a valid python name.
    def simulation_run(self, **kwparams):
        """Takes parameters as arguments, runs simulations as defined by events with them
        and returns the results as a series with MultiIndex[label, time].

        ParamSetError is raised if a parameter could not be set correctly,
        this exception indicates an error in the parameter configuration and should not be ignored.

        Note that kwparams needs keys that match the index of DataFrame self.param_specs.
        """

        if not set(kwparams.keys()) == set(self.param_spec.index):
            raise ValueError("Parameters need to match self.param_specs.")
        # set all parameters to the values defined by params.
        # param_specs defines where the parameter is located in the mike database
        for k in kwparams.keys():
            tab = self.param_spec.loc[k, "Table"]
            col = self.param_spec.loc[k, "Column"]
            muid = self.param_spec.loc[k, "Muid"]
            val = kwparams[k]
            self.dta.set_value(tab, muid, col, val)
            try:
                val_get = self.dta.get_field_values(tab, muid, [col])[0]
            except Exception as ex:
                raise ParamSetError(
                    f"Exception while trying to confirm parameter value. tab: {tab} col: {col} muid: {muid}"
                ) from ex
            if not val_get == val:
                raise ParamSetError(
                    f"""Value not correctly set. 
                            set {val}, got {val_get}.
                            tab: {tab}
                            col: {col}
                            muid: {muid}"""
                )
        if DELETE_RESFILE_BEFORE_SIM:
            for l in self.result_spec.index:
                self._delete_result_file(l)
        # Now run each simulation on its own thread. Map only returns when all simulations are finished.
        # console output is suppressed, because mikepluspy print far too much stuff for many runs.
        with open(os.devnull,mode="r+") as null_stream:
            with redirect_stdout(null_stream):
                finish_msgs = self.sim_executor.map(
                    self._single_sim_run, self.sim_longest_first
                )
        print(finish_msgs)
        for l in self.result_spec.index:
            assert self._exists_result_file(l), "Result file not found."
        return self.read_results()

    def read_results(self) -> pd.Series:
        """Reads all results according to event table. Reads each file at most once
        returns:
            a pandas Series with (label,Datetime) as index.
        """
        # collect results for each label in a dictionary
        result_dict = {}
        # go over all required files and read all results for that file in one go
        for result_file, events_file in self.result_spec.groupby("ResultFile"):
            # collect necessary reaches and nodes
            reaches = set(events_file["ResultReach"]) - {""}
            nodes = set(events_file["ResultNode"]) - {""}
            file_path: Path = self.result_folder / result_file
            params:dict[str,typing.Any] = {
                "file_path": str(file_path.absolute()),
            }
            # these parameters would make the reading faster, but currently do not work: the results are wrong.
            #if len(reaches) > 0:
            #    params["reaches"] = list(reaches)
            #if len(nodes) > 0:
            #    params["nodes"] = list(nodes)
            file_results = Res1D(**params).read_all()
            
            for label in events_file.index:
                column = events_file.at[label, "ResultColumn"]
                timestep = events_file.at[label, "Timestep"]
                start = events_file.at[label, "Start"]
                end = events_file.at[label, "End"]
                series = file_results[column]
                # Reindex to be comparable to results.
                # There should not be many missing values, otherwise issue a warning.
                # If you wand to use resampling on irregular spaced times, note that linear interpolation with pandas first throws away all
                # values where the index does not match to the frequency.
                full_idx = pd.date_range(start, end, freq=timestep)
                if len(full_idx.intersection(series.index)) < 0.95 * len(full_idx):
                    actual_timestep = series.index.to_series().diff().mean()
                    warn(
                        f"too few values in result for {label} in file '{file_path}'.\n"
                        + f"\tWant {start} to {end}, with timestep {timestep}.\n"
                        + f"\tGot  {series.index.min()} to {series.index.max()}, with timestep {actual_timestep}."
                        + "Timestep of result files may not match expectation ({timestep})."
                    )
                    print(series)
                series = series.reindex(full_idx)
                series.index.name = "Datetime"
                result_dict[label] = series
        return pd.concat(result_dict, names=["Label"])

    def _delete_result_file(self, label: str):
        """Deletes the file holding the result for label.
        Useful to make sure that no old values are read."""
        file_path: Path = self.result_folder / self.result_spec.at[label, "ResultFile"]
        file_path.unlink(True)

    def _exists_result_file(self, label: str):
        """Check if the file holding the result for label exists."""
        file_path: Path = self.result_folder / self.result_spec.at[label, "ResultFile"]
        return file_path.exists()


class ParamSetError(Exception):
    """Class that marks an Error where Parameters were not correctly set."""

    # class only needed to distinguish Error, no special behavior needed.
    pass
